import functools
from pyservice.action import Action
from pyservice.context import Context
from itertools import takewhile
from typing import List


class Organizer:
    class ContextFailed(Exception):
        def __init__(self, action: Action):
            self.action = action

    def __init__(self, actions: List[Action]):
        self.actions = actions

    def run(self, ctx: Context) -> Context:
        try:
            return functools.reduce(lambda _ctx, f: f(_ctx), self.actions, ctx)
        except Organizer.ContextFailed as e:
            # roll back the actions in reverse order
            actions_to_roll_back = self._find_actions_to_roll_back(
                e.action, self.actions
            )

            result = functools.reduce(self._execute_rollback, actions_to_roll_back, ctx)

            return ctx if result is None else result

    @staticmethod
    def _execute_rollback(ctx: Context, action: Action):
        return action(ctx)

    @staticmethod
    def _find_actions_to_roll_back(
        action: Action, actions: List[Action]
    ) -> List[Action]:

        actions_to_roll_back = [
            *takewhile(lambda a, x=action: a != x, actions)  # type: ignore
        ]
        actions_to_roll_back.reverse()
        return actions_to_roll_back



from typing import Any, Dict, Optional


class Context(dict):
    """ Wrapped dictionary carrying info between actions """

    def __init__(self, *arg, **kw):
        self.__success = True
        self.__skipped = False
        self.__message = None

    @property
    def is_failure(self) -> bool:
        return self.__success is False

    @property
    def is_success(self) -> bool:
        return self.__success

    @property
    def is_skipped(self) -> bool:
        return self.__skipped

    @property
    def message(self) -> Optional[str]:
        return self.__message

    def fail(self, msg: Optional[str] = None) -> None:
        if msg is not None:
            self.__message = msg

        self.__success = False

    def skip(self, msg: Optional[str] = None) -> None:
        if msg is not None:
            self.__message = msg

        self.__skipped = True

    @staticmethod
    def make(dict_value: Dict[str, Any] = {}) -> "Context":
        ctx = Context()
        ctx.update(dict_value)
        return ctx



from functools import wraps
from pyservice.context import Context
from typing import Callable, List, Optional

# Type declaration for Action
Action = Callable[[Context], Context]


class ExpectedKeyNotFoundError(Exception):
    pass


class PromisedKeyNotFoundError(Exception):
    pass


def _verify_expected_keys(expected_keys: List[str], ctx_keys: List[str]) -> None:
    if not expected_keys:
        return

    set_expected_keys = set(expected_keys)
    set_ctx_keys = set(ctx_keys)
    expected_diff = set_expected_keys - set_ctx_keys

    if expected_diff:
        raise ExpectedKeyNotFoundError(f"Missing keys: {list(expected_diff)}")


def _verify_promised_keys(promised_keys: List[str], ctx_keys: List[str]) -> None:
    if not promised_keys:
        return

    promised_diff = set(promised_keys) - set(ctx_keys)

    if promised_diff:
        raise PromisedKeyNotFoundError(
            f"Promised keys not found: {list(promised_diff)}"
        )


def action(
    expects: List[str] = [],
    promises: List[str] = [],
    rollback: Optional[Action] = None,
) -> Callable:
    def action_wrapper(f: Callable):
        @wraps(f)
        def decorated(ctx: Context) -> Context:
            _verify_expected_keys(expects, list(ctx.keys()))

            if ctx.is_skipped:
                return ctx

            if ctx.is_failure:
                if rollback:
                    return rollback(ctx)
                else:
                    return ctx

            result = f(ctx)

            _verify_promised_keys(promises, list(ctx.keys()))

            return result

        return decorated

    return action_wrapper



__version__ = "0.0.5"


from .action import (  # noqa: F401,E501
    action,
    ExpectedKeyNotFoundError,
    PromisedKeyNotFoundError,
)
from .context import Context  # noqa: F401,E501
from .organizer import Organizer  # noqa: F401,E501
